"""
    Copyright (C) 2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import atexit
import gzip
import time
import xml.etree.ElementTree as ET
from io import BytesIO

import requests

from .Addon import Addon
from ..versions import AddonVersion


class RateLimitedAdapter(requests.adapters.HTTPAdapter):
    def __init__(self, *args, retries=5, wait=None, **kwargs):
        self._last_send = None
        self._wait_time = wait
        max_retries = requests.adapters.Retry(
            total=retries,
            backoff_factor=wait or 10,
            status_forcelist={429, },
            allowed_methods=None,
        )
        kwargs.setdefault('max_retries', max_retries)
        super().__init__(*args, **kwargs)

    def send(self, *args, **kwargs):
        if self._wait_time and self._last_send:
            delta = time.time() - self._last_send
            if delta < self._wait_time:
                time.sleep(self._wait_time - delta)

        self._last_send = time.time()
        response = super().send(*args, **kwargs)
        status_code = getattr(response, 'status_code', None)
        if 300 <= status_code < 400:
            self._last_send = None
        return response


class Repository():
    # Recover from unreliable mirrors
    _session = requests.Session()
    _adapter = RateLimitedAdapter(retries=5, pool_maxsize=3, pool_block=True)
    _session.mount('http://', _adapter)
    _session.mount('https://', _adapter)
    atexit.register(_session.close)

    def __init__(self, version, path):
        super().__init__()
        self.version = version
        self.path = path

        try:
            response = self._session.get(path, timeout=(30, 30))
            response.raise_for_status()
        except requests.exceptions.RequestException:
            return
        content = response.content

        if path.endswith('.gz'):
            with gzip.open(BytesIO(content), 'rb') as xml_file:
                content = xml_file.read()

        self.addons = []
        tree = ET.fromstring(content)
        for addon in tree.findall("addon"):
            self.addons.append(Addon(addon))

    def __contains__(self, addonId):
        for addon in self.addons:
            if addon.id == addonId:
                return True
        return False

    def find(self, addonId):
        # multiple copies of the same addon might exist on the repository, however
        # kodi always uses the highest version available
        addon_instances = []
        for addon in self.addons:
            if addon.id == addonId:
                addon_instances.append(addon)

        if not addon_instances:
            return None

        # always return the highest version for the given addon id available in the repo
        addon_instances.sort(key=lambda addon: AddonVersion(addon.version), reverse=True)
        return addon_instances[0]

    def rdepends(self, addonId):
        rdepends = []
        for addon in self.addons:
            if addon.dependsOn(addonId):
                rdepends.append(addon)
        return rdepends



import unittest
from kodi_addon_checker.check_addon import get_all_repo_addons
from kodi_addon_checker.check_addon import start
from kodi_addon_checker.common import load_plugins
from kodi_addon_checker.record import Record
from kodi_addon_checker.reporter import ReportManager
from kodi_addon_checker.config import Config


class Args():
    PR = False
    allow_folder_id_mismatch = False
    branch = "krypton"


class TestCheckAddon(unittest.TestCase):
    """Integration tests for Start function present in check_addon.py"""

    def setUp(self):
        self.path = "script.test/"
        self.whitelist = ["INFO: Checking add-on script.test", "INFO: Created by mzfr", "INFO: This is a new addon"
                          "INFO: Image icon exists", "Icon dimensions are fine", "INFO: Image fanart exists",
                          "WARN: Complex entry point", "WARN: We found", "please check the logfile"]
        load_plugins()
        self.config = Config(self.path)
        ReportManager.enable(["array"])
        self.all_repo_addons = get_all_repo_addons()
        self.args = Args()

    def test_start(self):
        start(self.path, self.args, self.all_repo_addons, self.config)
        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]

        # Comparing the whitelist with the list of output we get from addon-checker tool
        for white_str in self.whitelist:
            for value in records:
                if white_str.lower() == value.lower():
                    break
            else:
                flag = False
        else:
            flag = True

        self.assertTrue(flag)



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import difflib
from lib2to3 import pgen2, refactor # pylint: disable=deprecated-module

from .common import relative_path
from .record import INFORMATION, PROBLEM, Record
from .report import Report
from .versions import KodiVersion


class KodiRefactoringTool(refactor.RefactoringTool):

    def __init__(self, report, log_level, *args, **kwargs):
        self.report = report
        self.log_level = log_level
        super().__init__(*args, **kwargs)

    def print_output(self, old_text, new_text, filename, equal):
        """
        Called with the old version, new version, and filepath of a
        refactored file.

            :old_text: old text of refactored file
            :new_text: new text of refactored file
            :filename: Path of the file
            :equal: Tells whether or not old is equal to new
        """

        if equal:
            return

        diff = ""
        for line in difflib.unified_diff(old_text.splitlines(), new_text.splitlines(),
                                         relative_path(filename), relative_path(filename),
                                         "(original)", "(refactored)", n=3, lineterm=""):
            diff += line + "\n"

        self.report.add(Record(self.log_level, relative_path(filename) + '\n' + diff[:-1]))


def check_py3_compatibility(report: Report, path: str, kodi_version: KodiVersion):
    """
     Checks compatibility of addons with python3
        :path: path to the addon
    """
    list_of_fixes = [
        'except',
        'exec',
        'ne',
        'print',
        'raise',
        'repr',
        'tuple_params',
    ]

    fixer_names = ['lib2to3.fixes.fix_' + fix for fix in list_of_fixes]

    rt = KodiRefactoringTool(report, PROBLEM, fixer_names, options={"print_function": True}, explicit=None)
    try:
        rt.refactor([path])
    except pgen2.parse.ParseError:
        rt = KodiRefactoringTool(report, PROBLEM, fixer_names, options=None, explicit=None)
        try:
            rt.refactor([path])
        except pgen2.parse.ParseError as pe:
            report.add(Record(PROBLEM, f"ParseError: {pe}"))
        except UnicodeDecodeError as ude:
            report.add(Record(PROBLEM, f"UnicodeDecodeError: {ude}"))
    except UnicodeDecodeError as ude:
        report.add(Record(PROBLEM, f"UnicodeDecodeError: {ude}"))

    if kodi_version >= KodiVersion("krypton"):
        list_of_fixes = [
                        'dict',
                        'filter',
                        'has_key',
                        'import',
                        'itertools',
                        'map',
                        'next',
                        'numliterals',
                        'renames',
                        'types',
                        'xrange',
                        'zip',
                        ]

        fixer_names = ['lib2to3.fixes.fix_' + fix for fix in list_of_fixes]

        rt = KodiRefactoringTool(report, INFORMATION, fixer_names, options={"print_function": True}, explicit=None)
        try:
            rt.refactor([path])
        except pgen2.parse.ParseError:
            rt = KodiRefactoringTool(report, INFORMATION, fixer_names, options=None, explicit=None)
            try:
                rt.refactor([path])
            except pgen2.parse.ParseError as pe:
                report.add(Record(INFORMATION, f"ParseError: {pe}"))
            except UnicodeDecodeError as ude:
                report.add(Record(PROBLEM, f"UnicodeDecodeError: {ude}"))
        except UnicodeDecodeError as ude:
            report.add(Record(PROBLEM, f"UnicodeDecodeError: {ude}"))



"""
    Copyright (C) 2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import xml.etree.ElementTree as ET
from ..versions import AddonVersion

class AddonDependency():
    def __init__(self, import_xml: ET.Element):
        super().__init__()
        self.id = import_xml.get('addon')
        self.version = None
        if import_xml.get('version') is not None:
            self.version = AddonVersion(import_xml.get('version'))
        self.optional = import_xml.get('optional', False)



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

__all__ = ["decorators", "has_transparency", "relative_path", "load_plugins"]

import importlib
import os
import pkgutil
import sys

REL_PATH = ""


def has_transparency(im):
    """Check the transparency(alpha layer) in the given image

        :im: PIL.Image object
    """
    try:
        if im.mode == "RGBA":
            alpha = im.split()[-1]
            listdata = list(alpha.getdata())
            first_transparent_pixel = next(x[0]
                                           for x in enumerate(listdata) if x[1] < 255)
            if first_transparent_pixel is not None:
                return True
        return False
    except StopIteration:
        return False


def load_plugins():
    """Load the reporter plugins
    """
    plugins_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "plugins")
    sys.path.append(plugins_dir)
    for _, package_name, _ in pkgutil.iter_modules([plugins_dir]):
        if "test_" not in package_name:
            importlib.import_module(package_name)


def relative_path(file_path):
    """Return the relative_path of the file
    """
    path_to_print = file_path[len(REL_PATH):]
    return f".{path_to_print}"



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import json
import os
import re
import xml.etree.ElementTree as ET

from . import handle_files
from .common import relative_path
from .common.decorators import posix_only
from .record import INFORMATION, PROBLEM, WARNING, Record
from .report import Report


def check_for_invalid_xml_files(report: Report, file_index: list):
    """check if any xml file present in the addon is invalid or not
        :file_index: A list having all the name and path of files in
                        addons
    """
    for file in file_index:
        if os.path.splitext(file["name"])[1].lower() == ".xml":
            xml_path = os.path.join(file["path"], file["name"])
            try:
                # Just try if we can successfully parse it
                ET.parse(xml_path)
            except ET.ParseError:
                report.add(Record(PROBLEM, f"Invalid xml found. {relative_path(xml_path)}"))


def check_for_invalid_json_files(report: Report, file_index: list):
    """ check if any json file present in the addon is invalid or not
        :file_index: A list having all the name and path of files in
                     addons
    """
    for file in file_index:
        if os.path.splitext(file["name"])[1].lower() == ".json":
            path = os.path.join(file["path"], file["name"])
            try:
                # Just try if we can successfully parse it
                with open(path, "r", encoding="utf8") as json_data:
                    json.load(json_data)
            except ValueError:
                report.add(Record(PROBLEM, f"Invalid json found. {relative_path(path)}"))


def check_addon_xml(report: Report, addon_path: str, parsed_xml, folder_id_mismatch: bool):
    """Check whether the addon.xml present in the addon is parseable or not
        :addon_path: path to the addon
        :parsed_xml: parsed tree for xml file
        :folder_id_mismatch: whether to allow folder and id mismatch
    """
    addon_xml_path = os.path.join(addon_path, "addon.xml")
    try:
        handle_files.addon_file_exists(report, addon_path, r"addon\.xml")

        report.add(Record(INFORMATION, f"Created by {parsed_xml.attrib.get('provider-name')}"))
        addon_xml_matches_folder(report, addon_path, parsed_xml, folder_id_mismatch)
    except ET.ParseError:
        report.add(Record(PROBLEM, f"Addon xml not valid, check xml. {relative_path(addon_xml_path)}"))

    return parsed_xml


def addon_xml_matches_folder(report: Report, addon_path: str, parsed_xml, folder_id_mismatch: bool):
    """Check if the name of the addon matches the folder in which the addon
    files are present
        :addon_path: path to the addon folder
        :addon_xml: parsed tree for xml file
        :folder_id_mismatch: whether to allow folder and id mismatch
    """
    addon_id = parsed_xml.attrib.get("id")
    if os.path.basename(os.path.normpath(addon_path)) == addon_id:
        report.add(Record(INFORMATION, "Addon id matches folder name"))
    else:
        if folder_id_mismatch:
            report.add(Record(INFORMATION, "Addon id and folder name does not match. "
                                           f"Ensure folder name is {addon_id} when submitting a PR "
                                           "to Kodi's official repository."))
        else:
            report.add(Record(PROBLEM, "Addon id and folder name does not match."))


def check_for_new_language_directory_structure(report: Report, addon_path: str, supported=True):
    """Check whether the language directory structure is new or not
        :addon_path: path to addon folder
        :supported: if we should error out in case the new language format is not supported
        by the respective kodiversion
    """
    language_path = os.path.join(addon_path, "resources", "language")
    if os.path.exists(language_path):
        dirs = next(os.walk(language_path))[1]
        for directory in dirs:
            if "resource.language." not in directory and supported:
                report.add(Record(
                    PROBLEM, "Using the old language directory structure in " \
                    f"{os.path.join(language_path, directory)}, please move to the new one."))
            elif "resource.language." in directory and not supported:
                report.add(Record(
                    PROBLEM, "Using the new language directory structure "\
                            f"in {os.path.join(language_path, directory)} " \
                             "for a Kodi version that does not support it. " \
                             "Please use the old language file structure or move the addon to" \
                             "an upper branch/kodi version."))


def check_file_whitelist(report: Report, file_index: list, addon_path: str):
    """check whether the files present in addon are in whitelist or not
        It ignores README.md and .gitignore file
        :file_index: list having names and path of all the files present in addon
        :addon_path: path to the addon folder
    """
    if ".module." in addon_path:
        report.add(Record(INFORMATION, "Module skipping whitelist"))
        return

    whitelist = (
        r"\.?(py|xml|gif|png|jpg|jpeg|md|txt|po|json|markdown|yml|"
        r"rst|ini|flv|wav|mp4|html|css|lst|pkla|g|template|in|cfg|xsd|directory|"
        r"help|list|mpeg|pls|info|ttf|xsp|theme|yaml|dict|crt|ico)?$"
    )

    for file in file_index:
        file_parts = file["name"].rsplit(".")
        if len(file_parts) > 1:
            file_ending = "." + file_parts[len(file_parts) - 1]
            if re.match(whitelist, file_ending, re.IGNORECASE) is None:
                filename = relative_path(os.path.join(file['path'], file['name']))
                report.add(Record(WARNING,
                                  f"Found non whitelisted file ending in filename {filename}"))


@posix_only
def check_file_permission(report: Report, file_index: list):
    """Check whether the files present in addon are marked executable
       or not
        :file_index: list having names and path of all the files present in addon
    """

    for file in file_index:
        file = os.path.join(file["path"], file["name"])
        if os.path.isfile(file) and os.access(str(file), os.X_OK):
            report.add(Record(PROBLEM, f"{relative_path(str(file))} is marked as stand-alone executable"))



import unittest
from os.path import abspath, dirname, join

from kodi_addon_checker.check_string import check_for_invalid_strings_po
from kodi_addon_checker.common import load_plugins
from kodi_addon_checker.common import relative_path
from kodi_addon_checker.record import Record
from kodi_addon_checker.reporter import ReportManager
from kodi_addon_checker.report import Report

HERE = abspath(dirname(__file__))


class TestPOFiles(unittest.TestCase):
    def setUp(self):
        self.path = join(HERE, "test_data", "PO_files")
        load_plugins()
        ReportManager.enable(["array"])
        self.report = Report("")
        self.report_matches = "ERROR: Invalid PO file"

    def test_check_for_invalid_strings_po_valid_file(self):
        ReportManager.getEnabledReporters()[0].reports = []

        path = join(self.path, "valid_file")
        language_path = join(path, "resources", "language", "resource.language.en_gb")

        file_index = [{"path": language_path, "name": "strings.po"}]

        expected = []
        check_for_invalid_strings_po(self.report, file_index)

        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        output = [s for s in records if s.startswith(self.report_matches)]

        self.assertListEqual(expected, output)

    def test_check_for_invalid_strings_po_missing_header(self):
        ReportManager.getEnabledReporters()[0].reports = []

        path = join(self.path, "missing_header")
        language_path = join(path, "resources", "language", "resource.language.en_gb")
        full_path = join(language_path, "strings.po")

        file_index = [{"path": language_path, "name": "strings.po"}]

        expected = [f'ERROR: Invalid PO file { relative_path(full_path) }:\n'
                    'Missing required header:\n'
                    '\tmsgid ""\n\tmsgstr ""']

        check_for_invalid_strings_po(self.report, file_index)

        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        output = [s for s in records if s.startswith(self.report_matches)]

        self.assertListEqual(expected, output)

    def test_check_for_invalid_strings_po_syntax_error(self):
        ReportManager.getEnabledReporters()[0].reports = []

        path = join(self.path, "syntax_error")
        language_path = join(path, "resources", "language", "resource.language.en_gb")
        full_path = join(language_path, "strings.po")

        file_index = [{"path": language_path, "name": "strings.po"}]

        expected = [f"ERROR: Invalid PO file { relative_path(full_path) }: "
                    "Syntax error on line 23"]

        check_for_invalid_strings_po(self.report, file_index)

        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        output = [s for s in records if s.startswith(self.report_matches)]

        self.assertListEqual(expected, output)

    def test_check_for_invalid_strings_po_encoding(self):
        ReportManager.getEnabledReporters()[0].reports = []

        path = join(self.path, "encoding")
        language_path = join(path, "resources", "language", "resource.language.en_gb")
        full_path = join(language_path, "strings.po")

        file_index = [{"path": language_path, "name": "strings.po"}]

        expected = [f"ERROR: Invalid PO file { relative_path(full_path) }: "
                    "File is not saved with UTF-8 encoding"]

        check_for_invalid_strings_po(self.report, file_index)

        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        output = [s for s in records if s.startswith(self.report_matches)]

        self.assertListEqual(expected, output)

    def test_check_for_invalid_strings_po_bom(self):
        ReportManager.getEnabledReporters()[0].reports = []

        path = join(self.path, "bom")
        language_path = join(path, "resources", "language", "resource.language.en_gb")
        full_path = join(language_path, "strings.po")

        file_index = [{"path": language_path, "name": "strings.po"}]

        expected = [f"ERROR: Invalid PO file { relative_path(full_path) }: "
                    "File contains BOM (byte order mark)"]

        check_for_invalid_strings_po(self.report, file_index)

        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        output = [s for s in records if s.startswith(self.report_matches)]

        self.assertListEqual(expected, output)

    def test_check_for_invalid_strings_po_empty(self):
        ReportManager.getEnabledReporters()[0].reports = []

        path = join(self.path, "empty")
        language_path = join(path, "resources", "language", "resource.language.en_gb")
        full_path = join(language_path, "strings.po")

        file_index = [{"path": language_path, "name": "strings.po"}]

        expected = [f"ERROR: Invalid PO file { relative_path(full_path) }: File is empty"]

        check_for_invalid_strings_po(self.report, file_index)

        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        output = [s for s in records if s.startswith(self.report_matches)]

        self.assertListEqual(expected, output)

    def test_check_for_invalid_strings_po_language_code(self):
        ReportManager.getEnabledReporters()[0].reports = []

        path = join(self.path, "path_check")
        language_path = join(path, "resources", "language")

        file_index = [{"path": join(language_path, "resource.language.testing"), "name": "strings.po"},
                      {"path": join(language_path, "resource.language.en_gb"), "name": "strings.po"}]

        expected = [
            "ERROR: PO file with invalid language code in the correct path: " \
            f"{ relative_path(join(language_path, 'resource.language.testing', 'strings.po')) }"
        ]

        check_for_invalid_strings_po(self.report, file_index)

        matches = (
            self.report_matches,
            "ERROR: PO file with invalid language code in the correct path"
        )

        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        output = [s for s in records if s.startswith(matches)]

        self.assertListEqual(expected, output)

    def test_check_for_invalid_strings_po_missing_en_gb(self):
        ReportManager.getEnabledReporters()[0].reports = []

        path = join(self.path, "path_check")
        language_path = join(path, "resources", "language")

        file_index = [{"path": join(language_path, "resource.language.en_us"), "name": "strings.po"}]

        expected = [
            "ERROR: Required default language 'en_gb' is not present."
        ]

        check_for_invalid_strings_po(self.report, file_index)

        matches = (
            self.report_matches,
            "ERROR: Required default language",
        )

        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        output = [s for s in records if s.startswith(matches)]
        self.assertListEqual(expected, output)



import os
import unittest
from os.path import abspath, dirname, join

from kodi_addon_checker.check_files import check_file_permission, check_file_whitelist
from kodi_addon_checker.handle_files import create_file_index

from kodi_addon_checker.common import load_plugins
from kodi_addon_checker.common import relative_path
from kodi_addon_checker.record import Record
from kodi_addon_checker.reporter import ReportManager
from kodi_addon_checker.report import Report


HERE = abspath(dirname(__file__))


class TestCheckFilePermission(unittest.TestCase):

    def setUp(self):
        load_plugins()
        ReportManager.enable(["array"])
        self.report = Report("")

    def test_check_file_permission_is_true(self):
        path = join(HERE, 'test_data', 'Executable_file')
        string = f"ERROR: { relative_path(join(path, 'file_permission.py')) } is marked as stand-alone executable"
        file_index = create_file_index(path)
        check_file_permission(self.report, file_index)
        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        flag = any(s == string for s in records)
        if os.name == "nt":
            self.assertFalse(flag)
        else:
            self.assertTrue(flag)

    def test_check_file_permission_is_None(self):
        path = join(HERE, 'test_data', 'Non-Executable_file')
        file_index = create_file_index(path)
        self.assertIsNone(check_file_permission(self.report, file_index))

    def test_gitignore(self):
        path = join(HERE, 'test_data', 'GitIgnore')
        string = f"WARN: Found non whitelisted file ending in filename { relative_path(join(path, '.gitignore')) }"
        file_index = create_file_index(path)
        check_file_whitelist(self.report, file_index, path)
        records = [Record.__str__(r) for r in ReportManager.getEnabledReporters()[0].reports]
        self.assertGreater(len(records), 0)
        self.assertEqual(records[-1], string)



import os
import setuptools

import kodi_addon_checker


with open('requirements.txt', 'r', encoding="utf8") as f:
    requirements = f.read().splitlines()

_ROOT = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(_ROOT, 'README.md'), 'r', encoding="utf8") as f:
    LONG_DESCRIPTION = f.read()

setuptools.setup(
    name="kodi-addon-checker",
    version=kodi_addon_checker.__version__,
    description="Check kodi addons or whole kodi repositories for errors and best practices.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Team Kodi",
    url="https://github.com/xbmc/addon-check",
    download_url="https://github.com/xbmc/addon-check/archive/master.zip",
    packages=setuptools.find_packages(exclude=['script.test', 'tests*']),
    package_data={'kodi_addon_checker': ['xml_schema/*.xsd']},
    install_requires=requirements,
    python_requires=">=3.5",
    setup_requires=['setuptools>=38.6.0'],
    entry_points={'console_scripts': [
        'kodi-addon-checker = kodi_addon_checker.__main__:main']},
    keywords='kodi add-on add-on_checker',
    classifiers=[
        "Operating System :: POSIX :: Linux",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Topic :: Utilities"
    ] + [f'Programming Language :: Python :: {x}' for x in '3 3.5 3.6 3.7 3.8 3.9 3.10 3.11'.split()]
)
